@@ -194,6 +194,7 @@ bpf_init(struct bpf_info *bpf)
 		MEMBER_OFFSET_INIT(bpf_prog_pages, "bpf_prog", "pages");
 		MEMBER_OFFSET_INIT(bpf_prog_aux_load_time, "bpf_prog_aux", "load_time");
 		MEMBER_OFFSET_INIT(bpf_prog_aux_user, "bpf_prog_aux", "user");
+		MEMBER_OFFSET_INIT(bpf_prog_aux_name, "bpf_prog_aux", "name");
 		MEMBER_OFFSET_INIT(bpf_map_key_size, "bpf_map", "key_size");
 		MEMBER_OFFSET_INIT(bpf_map_value_size, "bpf_map", "value_size");
 		MEMBER_OFFSET_INIT(bpf_map_max_entries, "bpf_map", "max_entries");
@@ -202,6 +203,13 @@ bpf_init(struct bpf_info *bpf)
 		MEMBER_OFFSET_INIT(bpf_map_user, "bpf_map", "user");
 		MEMBER_OFFSET_INIT(user_struct_uid, "user_struct", "uid");
 
+		/* Linux 5.3 */
+		MEMBER_OFFSET_INIT(bpf_map_memory, "bpf_map", "memory");
+		if (VALID_MEMBER(bpf_map_memory)) {
+			MEMBER_OFFSET_INIT(bpf_map_memory_pages, "bpf_map_memory", "pages");
+			MEMBER_OFFSET_INIT(bpf_map_memory_user, "bpf_map_memory", "user");
+		}
+
 		if (!bpf_type_size_init()) {
 			bpf->status = FALSE;
 			command_not_supported();
@@ -445,6 +453,17 @@ do_bpf(ulong flags, ulong prog_id, ulong map_id, int radix)
 			bpf_prog_gpl_compatible(buf1, (ulong)bpf->proglist[i].value);
 			fprintf(fp, "     GPL_COMPATIBLE: %s", buf1);
 
+			fprintf(fp, "  NAME: ");
+			if (VALID_MEMBER(bpf_prog_aux_name)) {
+				BCOPY(&bpf->bpf_prog_aux_buf[OFFSET(bpf_prog_aux_name)], buf1, 16);
+				buf1[16] = NULLCHAR;
+				if (strlen(buf1))
+					fprintf(fp, "\"%s\"", buf1);
+				else
+					fprintf(fp, "(unused)");
+			} else
+				fprintf(fp, "(unknown)");
+
 			fprintf(fp, "  UID: ");
 			if (VALID_MEMBER(bpf_prog_aux_user) && VALID_MEMBER(user_struct_uid)) {
 				user = ULONG(bpf->bpf_prog_aux_buf + OFFSET(bpf_prog_aux_user));
@@ -576,7 +595,11 @@ do_map_only:
 				fprintf(fp, "(unknown)");
 
 			fprintf(fp, "  MEMLOCK: ");
-			if (VALID_MEMBER(bpf_map_pages)) {
+			if (VALID_MEMBER(bpf_map_memory) && VALID_MEMBER(bpf_map_memory_pages)) {
+				map_pages = UINT(bpf->bpf_map_buf + OFFSET(bpf_map_memory)
+						+ OFFSET(bpf_map_memory_pages));
+				fprintf(fp, "%d\n", map_pages * PAGESIZE());
+			} else if (VALID_MEMBER(bpf_map_pages)) {
 				map_pages = UINT(bpf->bpf_map_buf + OFFSET(bpf_map_pages));
 				fprintf(fp, "%d\n", map_pages * PAGESIZE());
 			} else
@@ -594,8 +617,15 @@ do_map_only:
 				fprintf(fp, "(unknown)\n");
 
 			fprintf(fp, "  UID: ");
-			if (VALID_MEMBER(bpf_map_user) && VALID_MEMBER(user_struct_uid)) {
+			if (VALID_MEMBER(bpf_map_memory) && VALID_MEMBER(bpf_map_memory_user))
+				user = ULONG(bpf->bpf_map_buf + OFFSET(bpf_map_memory)
+						+ OFFSET(bpf_map_memory_user));
+			else if (VALID_MEMBER(bpf_map_user))
 				user = ULONG(bpf->bpf_map_buf + OFFSET(bpf_map_user));
+			else
+				user = 0;
+
+			if (user && VALID_MEMBER(user_struct_uid)) {
 				if (readmem(user + OFFSET(user_struct_uid), KVADDR, &uid, sizeof(uint), 
 				    "user_struct.uid", QUIET|RETURN_ON_ERROR))
 					fprintf(fp, "%d\n", uid);
