@@ -1277,8 +1277,7 @@ make_build_data(char *target)
         p = fgets(inbuf1, 79, fp1);
 
         p = fgets(inbuf2, 79, fp2);
-        p = strstr(inbuf2, ")");
-        p++;
+        p = strstr(inbuf2, " ");
         *p = '\0';
 
         p = fgets(inbuf3, 79, fp3);
@@ -1286,7 +1285,9 @@ make_build_data(char *target)
 	lower_case(target_data.program, progname);
 
 	fprintf(fp4, "char *build_command = \"%s\";\n", progname);
-        if (strlen(hostname))
+        if (getenv("SOURCE_DATE_EPOCH"))
+                fprintf(fp4, "char *build_data = \"reproducible build\";\n");
+        else if (strlen(hostname))
                 fprintf(fp4, "char *build_data = \"%s by %s on %s\";\n",
                         strip_linefeeds(inbuf1), inbuf2, hostname);
         else
@@ -1338,7 +1339,7 @@ make_spec_file(struct supported_gdb_version *sp)
 	printf("License: %s\n", sp->GPL);
 	printf("Group: Development/Debuggers\n");
 	printf("Source: %%{name}-%%{version}.tar.gz\n");
-	printf("URL: http://people.redhat.com/anderson\n");
+	printf("URL: https://github.com/crash-utility\n");
 	printf("Distribution: Linux 2.2 or greater\n");
 	printf("Vendor: Red Hat, Inc.\n");
 	printf("Packager: Dave Anderson <anderson@redhat.com>\n");
