@@ -31,9 +31,9 @@ static int per_cpu_buffer_sizes;
  */
 static int encapsulated_current_trace;
 /*
- * trace_buffer is supported
+ * array_buffer (trace_buffer pre v5.6) is supported
  */
-static int trace_buffer_available;
+static int array_buffer_available;
 /*
  * max_buffer is supported
  */
@@ -43,19 +43,24 @@ static int max_buffer_available;
  */
 static int multiple_instances_available;
 
+/*
+ * buffer_page has "real_end"
+ */
+static int buffer_page_real_end_available;
+
 #define koffset(struct, member) struct##_##member##_offset
 
 static int koffset(trace_array, current_trace);
-static int koffset(trace_array, trace_buffer);
+static int koffset(trace_array, array_buffer);
 static int koffset(trace_array, max_buffer);
-static int koffset(trace_buffer, buffer);
+static int koffset(array_buffer, buffer);
 static int koffset(trace_array, buffer);
 static int koffset(tracer, name);
 
-static int koffset(ring_buffer, pages);
-static int koffset(ring_buffer, flags);
-static int koffset(ring_buffer, cpus);
-static int koffset(ring_buffer, buffers);
+static int koffset(trace_buffer, pages);
+static int koffset(trace_buffer, flags);
+static int koffset(trace_buffer, cpus);
+static int koffset(trace_buffer, buffers);
 
 static int koffset(ring_buffer_per_cpu, cpu);
 static int koffset(ring_buffer_per_cpu, pages);
@@ -70,6 +75,7 @@ static int koffset(ring_buffer_per_cpu, entries);
 static int koffset(buffer_page, read);
 static int koffset(buffer_page, list);
 static int koffset(buffer_page, page);
+static int koffset(buffer_page, real_end);
 
 static int koffset(list_head, next);
 
@@ -111,9 +117,9 @@ static ulong max_tr_trace;
 
 struct trace_instance {
 	char name[NAME_MAX + 1];
-	ulong trace_buffer;
+	ulong array_buffer;
 	ulong max_buffer;
-	ulong ring_buffer;
+	ulong trace_buffer;
 	unsigned pages;
 	struct ring_buffer_per_cpu *buffers;
 
@@ -168,8 +174,7 @@ static int write_and_check(int fd, void *data, size_t size)
 
 static int init_offsets(void)
 {
-#define init_offset(struct, member) do {				\
-		koffset(struct, member) = MEMBER_OFFSET(#struct, #member);\
+#define check_offset(struct, member) do { \
 		if (koffset(struct, member) < 0) {			\
 			fprintf(fp, "failed to init the offset, struct:"\
 				#struct ", member:" #member);		\
@@ -178,12 +183,29 @@ static int init_offsets(void)
 		}							\
 	} while (0)
 
+#define init_offset(struct, member) do {				\
+		koffset(struct, member) = MEMBER_OFFSET(#struct, #member); \
+		check_offset(struct, member);				\
+	} while (0)
+
+#define init_offset_alternative(struct, member, alt_struct, alt_member) do {	\
+		koffset(struct, member) = MEMBER_OFFSET(#alt_struct, #alt_member); \
+		check_offset(struct, member);				\
+	} while (0)
+
 	if (encapsulated_current_trace)
 		init_offset(trace_array, current_trace);
 
-	if (trace_buffer_available) {
-		init_offset(trace_array, trace_buffer);
-		init_offset(trace_buffer, buffer);
+	if (array_buffer_available) {
+		if (MEMBER_EXISTS("trace_array", "array_buffer")) {
+			init_offset(trace_array, array_buffer);
+			init_offset(array_buffer, buffer);
+		} else {
+			init_offset_alternative(trace_array, array_buffer,
+						trace_array, trace_buffer);
+			init_offset_alternative(array_buffer, buffer,
+						trace_buffer, buffer);
+		}
 
 		if (max_buffer_available)
 			init_offset(trace_array, max_buffer);
@@ -203,11 +225,19 @@ static int init_offsets(void)
 	else if (kernel_symbol_exists("ring_buffer_read"))
 		gdb_set_crash_scope(symbol_value("ring_buffer_read"), "ring_buffer_read");
 
-	if (!per_cpu_buffer_sizes)
-		init_offset(ring_buffer, pages);
-	init_offset(ring_buffer, flags);
-	init_offset(ring_buffer, cpus);
-	init_offset(ring_buffer, buffers);
+	if (STREQ(MEMBER_TYPE_NAME("trace_buffer", "buffer"), "ring_buffer")) {
+		if (!per_cpu_buffer_sizes)
+			init_offset_alternative(trace_buffer, pages, ring_buffer, pages);
+		init_offset_alternative(trace_buffer, flags, ring_buffer, flags);
+		init_offset_alternative(trace_buffer, cpus, ring_buffer, cpus);
+		init_offset_alternative(trace_buffer, buffers, ring_buffer, buffers);
+	} else {
+		if (!per_cpu_buffer_sizes)
+			init_offset(trace_buffer, pages);
+		init_offset(trace_buffer, flags);
+		init_offset(trace_buffer, cpus);
+		init_offset(trace_buffer, buffers);
+	}
 
 	if (MEMBER_SIZE("ring_buffer_per_cpu", "pages") == sizeof(ulong)) {
 		lockless_ring_buffer = 1;
@@ -229,6 +259,7 @@ static int init_offsets(void)
 	init_offset(buffer_page, read);
 	init_offset(buffer_page, list);
 	init_offset(buffer_page, page);
+	init_offset(buffer_page, real_end);
 
 	init_offset(list_head, next);
 
@@ -264,10 +295,10 @@ static void print_offsets(void)
 	print_offset(trace_array, buffer);
 	print_offset(tracer, name);
 
-	print_offset(ring_buffer, pages);
-	print_offset(ring_buffer, flags);
-	print_offset(ring_buffer, cpus);
-	print_offset(ring_buffer, buffers);
+	print_offset(trace_buffer, pages);
+	print_offset(trace_buffer, flags);
+	print_offset(trace_buffer, cpus);
+	print_offset(trace_buffer, buffers);
 
 	print_offset(ring_buffer_per_cpu, cpu);
 	print_offset(ring_buffer_per_cpu, pages);
@@ -281,6 +312,7 @@ static void print_offsets(void)
 	print_offset(buffer_page, read);
 	print_offset(buffer_page, list);
 	print_offset(buffer_page, page);
+	print_offset(buffer_page, real_end);
 
 	print_offset(list_head, next);
 
@@ -295,6 +327,20 @@ static void print_offsets(void)
 #undef print_offset
 }
 
+static int buffer_page_has_data(ulong page)
+{
+	uint end;
+
+	if (!buffer_page_real_end_available)
+		return 1;
+
+	/* Only write pages with data in it */
+	read_value(end, page, buffer_page, real_end);
+	return end;
+out_fail:
+	return 0;
+}
+
 static int ftrace_init_pages(struct ring_buffer_per_cpu *cpu_buffer,
 		unsigned nr_pages)
 {
@@ -361,7 +407,8 @@ static int ftrace_init_pages(struct ring_buffer_per_cpu *cpu_buffer,
 
 	/* Setup linear pages */
 
-	cpu_buffer->linear_pages[count++] = cpu_buffer->reader_page;
+	if (buffer_page_has_data(cpu_buffer->reader_page))
+		cpu_buffer->linear_pages[count++] = cpu_buffer->reader_page;
 
 	if (cpu_buffer->reader_page == cpu_buffer->commit_page)
 		goto done;
@@ -406,12 +453,12 @@ static void ftrace_destroy_buffers(struct ring_buffer_per_cpu *buffers)
 }
 
 static int ftrace_init_buffers(struct ring_buffer_per_cpu *buffers,
-		ulong ring_buffer, unsigned pages)
+			       ulong trace_buffer, unsigned pages)
 {
 	int i;
 	ulong buffers_array;
 
-	read_value(buffers_array, ring_buffer, ring_buffer, buffers);
+	read_value(buffers_array, trace_buffer, trace_buffer, buffers);
 
 	for (i = 0; i < nr_cpu_ids; i++) {
 		if (!readmem(buffers_array + sizeof(ulong) * i, KVADDR,
@@ -463,32 +510,32 @@ out_fail:
 
 static int ftrace_init_trace(struct trace_instance *ti, ulong instance_addr)
 {
-	if (trace_buffer_available) {
-		ti->trace_buffer = instance_addr +
-				koffset(trace_array, trace_buffer);
-		read_value(ti->ring_buffer, ti->trace_buffer,
-				trace_buffer, buffer);
+	if (array_buffer_available) {
+		ti->array_buffer = instance_addr +
+				koffset(trace_array, array_buffer);
+		read_value(ti->trace_buffer, ti->array_buffer,
+			   array_buffer, buffer);
 
 		if (max_buffer_available) {
 			ti->max_buffer = instance_addr +
 					koffset(trace_array, max_buffer);
 			read_value(ti->max_tr_ring_buffer, ti->max_buffer,
-					trace_buffer, buffer);
+					array_buffer, buffer);
 		}
 	} else {
-		read_value(ti->ring_buffer, instance_addr, trace_array, buffer);
-		read_value(ti->pages, ti->ring_buffer, ring_buffer, pages);
+		read_value(ti->trace_buffer, instance_addr, trace_array, buffer);
+		read_value(ti->pages, ti->trace_buffer, trace_buffer, pages);
 
 		read_value(ti->max_tr_ring_buffer, max_tr_trace, trace_array, buffer);
 		if (ti->max_tr_ring_buffer)
-			read_value(ti->max_tr_pages, ti->max_tr_ring_buffer, ring_buffer, pages);
+			read_value(ti->max_tr_pages, ti->max_tr_ring_buffer, trace_buffer, pages);
 	}
 
 	ti->buffers = calloc(sizeof(*ti->buffers), nr_cpu_ids);
 	if (ti->buffers == NULL)
 		goto out_fail;
 
-	if (ftrace_init_buffers(ti->buffers, ti->ring_buffer,
+	if (ftrace_init_buffers(ti->buffers, ti->trace_buffer,
 			ti->pages) < 0)
 		goto out_fail;
 
@@ -647,6 +694,8 @@ static int ftrace_init(void)
 		ftrace_trace_arrays = sym_ftrace_trace_arrays->value;
 	}
 
+	if (MEMBER_EXISTS("buffer_page", "real_end"))
+		buffer_page_real_end_available = 1;
 
 	if (MEMBER_EXISTS("trace_array", "current_trace")) {
 		encapsulated_current_trace = 1;
@@ -658,8 +707,9 @@ static int ftrace_init(void)
 		current_trace = sym_current_trace->value;
 	}
 
-	if (MEMBER_EXISTS("trace_array", "trace_buffer")) {
-		trace_buffer_available = 1;
+	if (MEMBER_EXISTS("trace_array", "array_buffer") ||
+	    MEMBER_EXISTS("trace_array", "trace_buffer")) {
+		array_buffer_available = 1;
 
 		if (MEMBER_EXISTS("trace_array", "max_buffer"))
 			max_buffer_available = 1;
@@ -1809,7 +1859,7 @@ static void cmd_ftrace(void)
 static char *help_ftrace[] = {
 "trace",
 "show or dump the tracing info",
-"[ <show [-c <cpulist>] [-f [no]<flagname>]> | <dump [-sm] <dest-dir>> ]",
+"[ <show [-c <cpulist>] [-f [no]<flagname>]> | <dump [-sm] <dest-dir>> ] | <dump -t <trace.dat> ]",
 "trace",
 "    shows the current tracer and other informations.",
 "",
@@ -2184,7 +2234,8 @@ static int save_proc_kallsyms(int fd)
 			if (!strncmp(sp->name, "_MODULE_", strlen("_MODULE_")))
 				continue;
 
-			tmp_fprintf("%lx %c %s\t[%s]\n", sp->value, sp->type,
+			/* Currently sp->type for modules is not trusted */
+			tmp_fprintf("%lx %c %s\t[%s]\n", sp->value, 'm',
 					sp->name, lm->mod_name);
 		}
 	}
